/**
 * Copyright 2007 Nathanael Cottin
 */
package test;

import java.util.Collection;
import java.util.HashSet;

import sds.data.Message;
import sds.data.Reference;
import sds.error.SdsException;
import sds.more.Miscellaneous;
import sds.service.Daemon;
import sds.service.Manager;

public final class FeedbackManager extends Manager {

	private boolean[][]	neighbours;
	private Reference	daemon;

	public FeedbackManager(boolean[][] neighbours, Reference daemon) {
		super(false);
		this.neighbours = neighbours;
		this.daemon = daemon;
	}

	@Override
	public void execute() {
		System.out.println("Executing FeedbackManager: " + getReference());

		System.out.println("Registering " + neighbours.length + " instance(s)");
		Collection<Reference> invalidRef;
		try {
			invalidRef = register(FeedbackServer.class, neighbours.length,
				true, daemon);
		}
		catch (SdsException e) {
			processError(e);
			return;
		}

		System.out.println("Invalid references: " + invalidRef.size());
		for (Reference r : invalidRef) {
			System.out.println(r);
		}

		System.out.println("\n---------------------------------\n");
		Collection<Reference> ref = getDistributedObjectReferences(FeedbackServer.class);
		System.out.println("Registered references: " + ref.size());
		for (Reference r : ref) {
			System.out.println(r);
		}

		System.out.println("\n---------------------------------\n");
		Reference[] refs = Miscellaneous.toArray(ref);
		System.out.println("Sending neighbourhood to " + refs.length
			+ " distributed object(s)");
		for (int i = 0; i < refs.length; i++) {
			sendNeighbours(refs[i], getNeighbours(refs, neighbours[i]));
		}

		System.out.println("\n---------------------------------\n");
		System.out.println("Initializing algorithm");
		Message msg = new Message();
		msg.setTag(FeedbackTags.INIT_TAG);
		msg.setData("Message to deliver");
		deliver(msg, true, refs[0]);

		System.out.println("\n---------------------------------\n");
		System.out.println("End: killing all distant processes");
		killAll(false);
	}

	@Override
	public void processError(SdsException error) {
		System.out.println("FeedbackManager encountered the following error:");
		error.printStackTrace();
	}

	@Override
	public boolean recv(Message msg) throws SdsException {
		switch (msg.getTag()) {
			case FeedbackTags.NEIGHBOURS_TAG:
				// Response to blocking neighbours delivery
				return true;
			case FeedbackTags.END_TAG: {
				// Response to blocking distributed algorithm execution
				return true;
			}
			default:
				return false;
		}
	}

	private void sendNeighbours(Reference recipient, Collection<Reference> n) {
		Message msg = new Message();
		msg.setTag(FeedbackTags.NEIGHBOURS_TAG);
		msg.setData(n);
		deliver(msg, true, recipient);
	}

	private Collection<Reference> getNeighbours(Reference[] refs, boolean[] n) {
		Collection<Reference> res = new HashSet<Reference>(n.length);
		for (int i = 0; i < n.length; i++) {
			if (n[i]) {
				res.add(refs[i]);
			}
		}

		return res;
	}

	@Override
	public void terminateServer() {
	}

	public static void main(String[] args) {
		try {
			// Graph incidence matrix with 3 nodes
			boolean[][] neighbours = { { false, true, false },
					{ true, false, true }, { true, false, false } };

			// Launch daemon on localhost and port 666, and set distributed
			// objects ports from 1001 to 1004
			Daemon daemon = new Daemon(false, 1001, 1001 + neighbours.length,
				Daemon.DEFAULT_FIRST_ID);
			daemon.run(666, true);

			Reference daemonReference = daemon.getReference();
			FeedbackManager mgr = new FeedbackManager(neighbours,
				daemonReference);
			mgr.registerDaemons(daemonReference);
			mgr.run(1000);

			// Also terminate manager and daemon to leave properly
			mgr.terminate();
			daemon.terminate();
		}
		catch (Exception e) {
			e.printStackTrace();
		}
	}
}
